# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import subprocess
from subprocess import PIPE
from typing import Any, Union

_PATH = str


class Commands(object):
    '''Defines execution for the standard commands.'''

    def __init__(self, executable: _PATH = 'default') -> None:
        '''Creates a new instance of the Commands.

        Args:
            executable_path: Path to the AndroidDriver. On the Windows platform, the best choice is default.
        '''

        _default_path = os.path.join(
            os.path.dirname(__file__), 'executable', 'adb.exe')

        if executable == 'default':
            self.path = _default_path
        elif executable.endswith('adb.exe'):
            if not os.path.isfile(executable):
                raise FileNotFoundError(f'{self.path!r} does not exist.')
            self.path = executable
        elif executable in ['adb', 'adb.exe']:
            PATH = os.environ['PATH']
            if not ('adb' in PATH or 'android' in PATH or 'platform-tools' in PATH):
                raise EnvironmentError('PATH does not exist.')
            self.path = executable
        else:
            self.path = _default_path

    def _build_cmd(self, args: Union[list, tuple]) -> str:
        '''Build command.'''
        cmd = [self.path]
        cmd.extend(args)
        return cmd

    def execute(self, *, args: Union[list, tuple], options: dict) -> tuple:
        '''Execute command.'''
        cmd = self._build_cmd(args)
        process = subprocess.Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE,
                                   encoding='utf-8', shell=options.get('shell', False), env=options.get('env'))
        return process



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, Tuple

from .keys import Keys


class Elements(object):
    '''Represents a element.'''

    def __init__(self, parent, node: dict, key: Keys, value: str, coord: List[int], click_point: Tuple[int]) -> None:
        self._parent = parent
        self._node = node
        self._key = key
        self._value = value
        self._coord = coord
        self._click_point = click_point

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} (element="{2}: {3}", coord="{4}")>'.format(type(self), self._key, self._value, self._coord)

    @property
    def text(self) -> str:
        """The text of the element."""
        return self._node['text']

    @property
    def resource_id(self) -> str:
        """The resource-id of the element."""
        return self._node['resource-id']

    @property
    def class_(self) -> str:
        """The class of the element."""
        return self._node['class']

    @property
    def package(self) -> str:
        """The package of the element."""
        return self._node['package']

    @property
    def content_desc(self) -> str:
        """The content-desc of the element."""
        return self._node['content-desc']

    @property
    def checkable(self) -> bool:
        """The checkable of the element."""
        return self._node['checkable']

    @property
    def bounds(self) -> List[int]:
        """The bounds of the element."""
        return self._coord

    @property
    def coord(self) -> List[int]:
        """The coord of the element."""
        return self._coord

    def click(self) -> None:
        """Clicks the element."""
        self._parent.click(*self._click_point)

    def clear(self) -> None:
        """Clears the text if it's a text entry element."""
        self.click()
        for i in self.text:
            self._parent.send_keyevents(Keys.DEL)

    def send_keys(self, text: str = 'cerium') -> None:
        '''Simulates typing keys.'''
        self.click()
        self._parent.send_keys(text)

    def is_selected(self) -> bool:
        """Returns whether the element is selected.

        Can be used to check if a checkbox or radio button is selected.
        """
        return self._node['selected']

    def is_enabled(self) -> bool:
        """Returns whether the element is enabled."""
        return self._node['enabled']

    def is_checked(self) -> bool:
        """Returns whether the element is checked."""
        return self._node['checked']

    def is_clickable(self) -> bool:
        """Returns whether the element is clickable."""
        return self._node['clickable']

    def is_focusable(self) -> bool:
        """Returns whether the element is focusable."""
        return self._node['focusable']

    def is_focused(self) -> bool:
        """Returns whether the element is focused."""
        return self._node['focused']

    def is_scrollable(self) -> bool:
        """Returns whether the element is scrollable."""
        return self._node['scrollable']

    def is_long_clickable(self) -> bool:
        """Returns whether the element is long-clickable."""
        return self._node['long-clickable']

    def is_password(self) -> bool:
        """Returns whether the element is password."""
        return self._node['password']



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, Union

from .commands import _PATH, Commands
from .utils import free_port, is_connectable, merge_dict


class BaseService(Commands):
    '''Object that manages the starting and stopping of the AndroidDriver.'''

    def __init__(self, executable: _PATH = 'default', port: Union[int, str] = 5037, env: Dict = None) -> None:
        super(BaseService, self).__init__(executable)
        self.port = port
        if self.port == 0:
            self.port = free_port()
        self.options = {'env': env}

    @property
    def service_tcp(self) -> str:
        '''Gets the TCP of the Service.'''
        return f"tcp://localhost:{self.port}"

    def service_args(self) -> NotImplemented:
        raise NotImplemented(
            "This method needs to be implemented in a sub class")

    def _build_cmd(self, args: Union[list, tuple]) -> str:
        '''Build command.'''
        cmd = [self.path]
        cmd.extend(self.service_args())
        cmd.extend(args)
        return cmd

    def _execute(self, *args: str, **kwargs: Any) -> tuple:
        '''Execute command.'''
        return self.execute(args=args, options=merge_dict(self.options, kwargs)).communicate()

    def start(self) -> None:
        '''Starts the Service.'''
        self._execute('start-server')

    def stop(self) -> None:
        '''Stops the service.'''
        self._execute('kill-server')

    def restart(self) -> None:
        '''Restart the server if it is running.'''
        self.stop()
        self.start()

    def version(self) -> str:
        '''Show the version number of Android Debug Bridge.'''
        output, _ = self._execute('version')
        return output.splitlines()[0].split()[-1]

    def devices(self) -> list:
        '''List connected devices.'''
        output, _ = self._execute('devices')
        return output.split()[4::2]

    def devices_l(self) -> Dict:
        '''List connected devices (-l for long output).'''
        output, _ = self._execute('devices', '-l')
        devices = output.split()[4::6]
        models = output.split()[7::6]
        return dict(zip(devices, models))

    def connect(self, host: str = '192.168.0.3', port: Union[int, str] = 5555) -> None:
        '''Connect to a device via TCP/IP directly.'''
        self.device_sn = f'{host}:{port}'
        if not is_connectable(host, port):
            raise ConnectionError(f'Cannot connect to {self.device_sn}.')
        self._execute('connect', self.device_sn)

    def __disconnect(self, host: str = '192.168.0.3', port: Union[int, str] = 5555) -> None:
        '''Disconnect from given TCP/IP device [default port=5555].'''
        self.device_sn = None
        self._execute('disconnect', f'{host}:{port}')

    def __disconnect_all(self) -> None:
        '''Disconnect all.'''
        self.device_sn = None
        self._execute('disconnect')

    def get_state(self) -> str:
        '''offline | bootloader | device'''
        output, error = self._execute('get-state')
        if error:
            raise DeviceConnectionException(
                'No devices are connected. Please connect the device with USB or via WLAN and turn on the USB debugging option.')
        return output.strip()


class Service(BaseService):
    '''Object that manages the starting and stopping of the AndroidDriver.'''

    def __init__(self, executable_path: _PATH = 'default', port: Union[int, str] = 5037, env: Dict = None, service_args: Union[list, tuple] = None) -> None:
        '''Creates a new instance of the Service.

        Args:
            executable_path: Path to the AndroidDriver.
            port: Port the service is running on.
            env: Environment variables.
            service_args: List of args to pass to the androiddriver service.
        '''

        self._service_args = service_args or []

        super(Service, self).__init__(executable_path, port=port, env=env)

    def service_args(self) -> str:
        '''Parameters when starting the service.'''
        return ['-P', str(self.port)] + self._service_args
